@@ -1,6 +1,6 @@
 # coding: utf-8
 
-from __future__ import absolute_import, unicode_literals
+from __future__ import absolute_import, unicode_literals, print_function
 
 import sys
 import os
@@ -35,9 +35,14 @@ from ruamel.yaml.constructor import (
 from ruamel.yaml.loader import Loader as UnsafeLoader
 
 if False:  # MYPY
-    from typing import List, Set, Dict, Union, Any  # NOQA
+    from typing import List, Set, Dict, Union, Any, Callable  # NOQA
     from ruamel.yaml.compat import StreamType, StreamTextType, VersionType  # NOQA
 
+    if PY3:
+        from pathlib import Path
+    else:
+        Path = Any
+
 try:
     from _ruamel_yaml import CParser, CEmitter  # type: ignore
 except:  # NOQA
@@ -53,8 +58,10 @@ enforce = object()
 
 
 class YAML(object):
-    def __init__(self, _kw=enforce, typ=None, pure=False, plug_ins=None):
-        # type: (Any, Any, Any, Any) -> None
+    def __init__(
+        self, _kw=enforce, typ=None, pure=False, output=None, plug_ins=None  # input=None,
+    ):
+        # type: (Any, Optional[Text], Any, Any, Any) -> None
         """
         _kw: not used, forces keyword arguments in 2.7 (in 3 you can do (*, safe_load=..)
         typ: 'rt'/None -> RoundTripLoader/RoundTripDumper,  (default)
@@ -62,6 +69,7 @@ class YAML(object):
              'unsafe'  -> normal/unsafe Loader/Dumper
              'base'    -> baseloader
         pure: if True only use Python modules
+        input/output: needed to work as context manager
         plug_ins: a list of plug-in files
         """
         if _kw is not enforce:
@@ -72,6 +80,11 @@ class YAML(object):
 
         self.typ = 'rt' if typ is None else typ
         self.pure = pure
+
+        # self._input = input
+        self._output = output
+        self._context_manager = None  # type: Any
+
         self.plug_ins = []  # type: List[Any]
         for pu in ([] if plug_ins is None else plug_ins) + self.official_plug_ins():
             file_name = pu.replace(os.sep, '.')
@@ -180,7 +193,7 @@ class YAML(object):
                     # else:
                     setattr(self, attr, CParser(self._stream))
                     # self._parser = self._composer = self
-                    # print('scanner', self.loader.scanner)
+                    # nprint('scanner', self.loader.scanner)
 
         return getattr(self, attr)
 
@@ -278,8 +291,19 @@ class YAML(object):
 
     # separate output resolver?
 
+    # def load(self, stream=None):
+    #     if self._context_manager:
+    #        if not self._input:
+    #             raise TypeError("Missing input stream while dumping from context manager")
+    #         for data in self._context_manager.load():
+    #             yield data
+    #         return
+    #     if stream is None:
+    #         raise TypeError("Need a stream argument when not loading from context manager")
+    #     return self.load_one(stream)
+
     def load(self, stream):
-        # type: (StreamTextType) -> Any
+        # type: (Union[Path, StreamTextType]) -> Any
         """
         at this point you either have the non-pure Parser (which has its own reader and
         scanner) or you have the pure Parser.
@@ -306,7 +330,7 @@ class YAML(object):
                 pass
 
     def load_all(self, stream, _kw=enforce):  # , skip=None):
-        # type: (StreamTextType, Any) -> Any
+        # type: (Union[Path, StreamTextType], Any) -> Any
         if _kw is not enforce:
             raise TypeError(
                 '{}.__init__() takes no positional argument but at least '
@@ -370,7 +394,7 @@ class YAML(object):
 
                 class XLoader(self.Parser, self.Constructor, rslvr):  # type: ignore
                     def __init__(selfx, stream, version=None, preserve_quotes=None):
-                        # type: (StreamTextType, Union[None, VersionType], Union[None, bool]) -> None  # NOQA
+                        # type: (StreamTextType, Optional[VersionType], Optional[bool]) -> None  # NOQA
                         CParser.__init__(selfx, stream)
                         selfx._parser = selfx._composer = selfx
                         self.Constructor.__init__(selfx, loader=selfx)
@@ -382,12 +406,46 @@ class YAML(object):
                 return loader, loader
         return self.constructor, self.parser
 
-    def dump(self, data, stream, _kw=enforce, transform=None):
-        # type: (Any, StreamType, Any, Any) -> Any
-        return self.dump_all([data], stream, _kw, transform=transform)
+    def dump(self, data, stream=None, _kw=enforce, transform=None):
+        # type: (Any, Union[Path, StreamType], Any, Any) -> Any
+        if self._context_manager:
+            if not self._output:
+                raise TypeError('Missing output stream while dumping from context manager')
+            if _kw is not enforce:
+                raise TypeError(
+                    '{}.dump() takes one positional argument but at least '
+                    'two were given ({!r})'.format(self.__class__.__name__, _kw)
+                )
+            if transform is not None:
+                raise TypeError(
+                    '{}.dump() in the context manager cannot have transform keyword '
+                    ''.format(self.__class__.__name__)
+                )
+            self._context_manager.dump(data)
+        else:  # old style
+            if stream is None:
+                raise TypeError('Need a stream argument when not dumping from context manager')
+            return self.dump_all([data], stream, _kw, transform=transform)
 
     def dump_all(self, documents, stream, _kw=enforce, transform=None):
-        # type: (Any, StreamType, Any, Any) -> Any
+        # type: (Any, Union[Path, StreamType], Any, Any) -> Any
+        if self._context_manager:
+            raise NotImplementedError
+        if _kw is not enforce:
+            raise TypeError(
+                '{}.dump(_all) takes two positional argument but at least '
+                'three were given ({!r})'.format(self.__class__.__name__, _kw)
+            )
+        self._output = stream
+        self._context_manager = YAMLContextManager(self, transform=transform)
+        for data in documents:
+            self._context_manager.dump(data)
+        self._context_manager.teardown_output()
+        self._output = None
+        self._context_manager = None
+
+    def Xdump_all(self, documents, stream, _kw=enforce, transform=None):
+        # type: (Any, Union[Path, StreamType], Any, Any) -> Any
         """
         Serialize a sequence of Python objects into a YAML stream.
         """
@@ -420,7 +478,7 @@ class YAML(object):
                 try:
                     self.representer.represent(data)
                 except AttributeError:
-                    # print(dir(dumper._representer))
+                    # nprint(dir(dumper._representer))
                     raise
             self.serializer.close()
         finally:
@@ -484,7 +542,7 @@ class YAML(object):
                 top_level_colon_align=None,
                 prefix_colon=None,
             ):
-                # type: (StreamType, Any, Any, Any, Union[None, bool], Union[None, int], Union[None, int], Union[None, bool], Any, Any, Union[None, bool], Union[None, bool], Any, Any, Any, Any, Any) -> None   # NOQA
+                # type: (StreamType, Any, Any, Any, Optional[bool], Optional[int], Optional[int], Optional[bool], Any, Any, Optional[bool], Optional[bool], Any, Any, Any, Any, Any) -> None   # NOQA
                 CEmitter.__init__(
                     selfx,
                     stream,
@@ -581,6 +639,41 @@ class YAML(object):
             self.constructor.add_constructor(tag, f_y)
         return cls
 
+    def parse(self, stream):
+        # type: (StreamTextType) -> Any
+        """
+        Parse a YAML stream and produce parsing events.
+        """
+        _, parser = self.get_constructor_parser(stream)
+        try:
+            while parser.check_event():
+                yield parser.get_event()
+        finally:
+            parser.dispose()
+            try:
+                self._reader.reset_reader()  # type: ignore
+            except AttributeError:
+                pass
+            try:
+                self._scanner.reset_scanner()  # type: ignore
+            except AttributeError:
+                pass
+
+    # ### context manager
+
+    def __enter__(self):
+        # type: () -> Any
+        self._context_manager = YAMLContextManager(self)
+        return self
+
+    def __exit__(self, typ, value, traceback):
+        # type: (Any, Any, Any) -> None
+        if typ:
+            nprint('typ', typ)
+        self._context_manager.teardown_output()
+        # self._context_manager.teardown_input()
+        self._context_manager = None
+
     # ### backwards compatibility
     def _indent(self, mapping=None, sequence=None, offset=None):
         # type: (Any, Any, Any) -> None
@@ -612,6 +705,113 @@ class YAML(object):
         self.sequence_dash_offset = val
 
 
+class YAMLContextManager(object):
+    def __init__(self, yaml, transform=None):
+        # type: (Any, Optional[Callable]) -> None
+        self._yaml = yaml
+        self._output_inited = False
+        self._output_path = None
+        self._output = self._yaml._output
+        self._transform = transform
+
+        # self._input_inited = False
+        # self._input = input
+        # self._input_path = None
+        # self._transform = yaml.transform
+        # self._fstream = None
+
+        if not hasattr(self._output, 'write') and hasattr(self._output, 'open'):
+            # pathlib.Path() instance, open with the same mode
+            self._output_path = self._output
+            self._output = self._output_path.open('w')
+
+        # if not hasattr(self._stream, 'write') and hasattr(stream, 'open'):
+        # if not hasattr(self._input, 'read') and hasattr(self._input, 'open'):
+        #    # pathlib.Path() instance, open with the same mode
+        #    self._input_path = self._input
+        #    self._input = self._input_path.open('r')
+
+        if self._transform is not None:
+            self._fstream = self._output
+            if self._yaml.encoding is None:
+                self._output = StringIO()
+            else:
+                self._output = BytesIO()
+
+    def teardown_output(self):
+        # type: () -> None
+        if self._output_inited:
+            self._yaml.serializer.close()
+        else:
+            return
+        try:
+            self._yaml.emitter.dispose()
+        except AttributeError:
+            raise
+            # self.dumper.dispose()  # cyaml
+        try:
+            delattr(self._yaml, '_serializer')
+            delattr(self._yaml, '_emitter')
+        except AttributeError:
+            raise
+        if self._transform:
+            val = self._output.getvalue()
+            if self._yaml.encoding:
+                val = val.decode(self._yaml.encoding)
+            if self._fstream is None:
+                self._transform(val)
+            else:
+                self._fstream.write(self._transform(val))
+                self._fstream.flush()
+                self._output = self._fstream  # maybe not necessary
+        if self._output_path is not None:
+            self._output.close()
+
+    def init_output(self, first_data):
+        # type: (Any) -> None
+        if self._yaml.top_level_colon_align is True:
+            tlca = max([len(str(x)) for x in first_data])  # type: Any
+        else:
+            tlca = self._yaml.top_level_colon_align
+        self._yaml.get_serializer_representer_emitter(self._output, tlca)
+        self._yaml.serializer.open()
+        self._output_inited = True
+
+    def dump(self, data):
+        # type: (Any) -> None
+        if not self._output_inited:
+            self.init_output(data)
+        try:
+            self._yaml.representer.represent(data)
+        except AttributeError:
+            # nprint(dir(dumper._representer))
+            raise
+
+    # def teardown_input(self):
+    #     pass
+    #
+    # def init_input(self):
+    #     # set the constructor and parser on YAML() instance
+    #     self._yaml.get_constructor_parser(stream)
+    #
+    # def load(self):
+    #     if not self._input_inited:
+    #         self.init_input()
+    #     try:
+    #         while self._yaml.constructor.check_data():
+    #             yield self._yaml.constructor.get_data()
+    #     finally:
+    #         parser.dispose()
+    #         try:
+    #             self._reader.reset_reader()  # type: ignore
+    #         except AttributeError:
+    #             pass
+    #         try:
+    #             self._scanner.reset_scanner()  # type: ignore
+    #         except AttributeError:
+    #             pass
+
+
 def yaml_object(yml):
     # type: (Any) -> Any
     """ decorator for classes that needs to dump/load objects
@@ -706,7 +906,7 @@ def compose_all(stream, Loader=Loader):
 
 
 def load(stream, Loader=None, version=None, preserve_quotes=None):
-    # type: (StreamTextType, Any, Union[None, VersionType], Any) -> Any
+    # type: (StreamTextType, Any, Optional[VersionType], Any) -> Any
     """
     Parse the first YAML document in a stream
     and produce the corresponding Python object.
@@ -719,10 +919,18 @@ def load(stream, Loader=None, version=None, preserve_quotes=None):
         return loader._constructor.get_single_data()
     finally:
         loader._parser.dispose()
+        try:
+            loader._reader.reset_reader()
+        except AttributeError:
+            pass
+        try:
+            loader._scanner.reset_scanner()
+        except AttributeError:
+            pass
 
 
 def load_all(stream, Loader=None, version=None, preserve_quotes=None):
-    # type: (Union[None, StreamTextType], Any, Union[None, VersionType], Union[None, bool]) -> Any  # NOQA
+    # type: (Optional[StreamTextType], Any, Optional[VersionType], Optional[bool]) -> Any  # NOQA
     """
     Parse all YAML documents in a stream
     and produce corresponding Python objects.
@@ -736,10 +944,18 @@ def load_all(stream, Loader=None, version=None, preserve_quotes=None):
             yield loader._constructor.get_data()
     finally:
         loader._parser.dispose()
+        try:
+            loader._reader.reset_reader()
+        except AttributeError:
+            pass
+        try:
+            loader._scanner.reset_scanner()
+        except AttributeError:
+            pass
 
 
 def safe_load(stream, version=None):
-    # type: (StreamTextType, Union[None, VersionType]) -> Any
+    # type: (StreamTextType, Optional[VersionType]) -> Any
     """
     Parse the first YAML document in a stream
     and produce the corresponding Python object.
@@ -749,7 +965,7 @@ def safe_load(stream, version=None):
 
 
 def safe_load_all(stream, version=None):
-    # type: (StreamTextType, Union[None, VersionType]) -> Any
+    # type: (StreamTextType, Optional[VersionType]) -> Any
     """
     Parse all YAML documents in a stream
     and produce corresponding Python objects.
@@ -759,7 +975,7 @@ def safe_load_all(stream, version=None):
 
 
 def round_trip_load(stream, version=None, preserve_quotes=None):
-    # type: (StreamTextType, Union[None, VersionType], Union[None, bool]) -> Any
+    # type: (StreamTextType, Optional[VersionType], Optional[bool]) -> Any
     """
     Parse the first YAML document in a stream
     and produce the corresponding Python object.
@@ -769,7 +985,7 @@ def round_trip_load(stream, version=None, preserve_quotes=None):
 
 
 def round_trip_load_all(stream, version=None, preserve_quotes=None):
-    # type: (StreamTextType, Union[None, VersionType], Union[None, bool]) -> Any
+    # type: (StreamTextType, Optional[VersionType], Optional[bool]) -> Any
     """
     Parse all YAML documents in a stream
     and produce corresponding Python objects.
@@ -788,7 +1004,7 @@ def emit(
     allow_unicode=None,
     line_break=None,
 ):
-    # type: (Any, Union[None, StreamType], Any, Union[None, bool], Union[int, None], Union[None, int], Union[None, bool], Any) -> Any  # NOQA
+    # type: (Any, Optional[StreamType], Any, Optional[bool], Union[int, None], Optional[int], Optional[bool], Any) -> Any  # NOQA
     """
     Emit YAML parsing events into a stream.
     If stream is None, return the produced string instead.
@@ -836,7 +1052,7 @@ def serialize_all(
     version=None,
     tags=None,
 ):
-    # type: (Any, Union[None, StreamType], Any, Any, Union[None, int], Union[None, int], Union[None, bool], Any, Any, Union[None, bool], Union[None, bool], Union[None, VersionType], Any) -> Any # NOQA
+    # type: (Any, Optional[StreamType], Any, Any, Optional[int], Optional[int], Optional[bool], Any, Any, Optional[bool], Optional[bool], Optional[VersionType], Any) -> Any # NOQA
     """
     Serialize a sequence of representation trees into a YAML stream.
     If stream is None, return the produced string instead.
@@ -877,7 +1093,7 @@ def serialize_all(
 
 
 def serialize(node, stream=None, Dumper=Dumper, **kwds):
-    # type: (Any, Union[None, StreamType], Any, Any) -> Any
+    # type: (Any, Optional[StreamType], Any, Any) -> Any
     """
     Serialize a representation tree into a YAML stream.
     If stream is None, return the produced string instead.
@@ -905,7 +1121,7 @@ def dump_all(
     top_level_colon_align=None,
     prefix_colon=None,
 ):
-    # type: (Any, Union[None, StreamType], Any, Any, Any, Union[None, bool], Union[None, int], Union[None, int], Union[None, bool], Any, Any, Union[None, bool], Union[None, bool], Any, Any, Any, Any, Any) -> Union[None, str]   # NOQA
+    # type: (Any, Optional[StreamType], Any, Any, Any, Optional[bool], Optional[int], Optional[int], Optional[bool], Any, Any, Optional[bool], Optional[bool], Any, Any, Any, Any, Any) -> Optional[str]   # NOQA
     """
     Serialize a sequence of Python objects into a YAML stream.
     If stream is None, return the produced string instead.
@@ -943,7 +1159,7 @@ def dump_all(
             try:
                 dumper._representer.represent(data)
             except AttributeError:
-                # print(dir(dumper._representer))
+                # nprint(dir(dumper._representer))
                 raise
         dumper._serializer.close()
     finally:
@@ -975,7 +1191,7 @@ def dump(
     tags=None,
     block_seq_indent=None,
 ):
-    # type: (Any, Union[None, StreamType], Any, Any, Any, Union[None, bool], Union[None, int], Union[None, int], Union[None, bool], Any, Any, Union[None, bool], Union[None, bool], Union[None, VersionType], Any, Any) -> Union[None, str]   # NOQA
+    # type: (Any, Optional[StreamType], Any, Any, Any, Optional[bool], Optional[int], Optional[int], Optional[bool], Any, Any, Optional[bool], Optional[bool], Optional[VersionType], Any, Any) -> Optional[str]   # NOQA
     """
     Serialize a Python object into a YAML stream.
     If stream is None, return the produced string instead.
@@ -1004,7 +1220,7 @@ def dump(
 
 
 def safe_dump_all(documents, stream=None, **kwds):
-    # type: (Any, Union[None, StreamType], Any) -> Union[None, str]
+    # type: (Any, Optional[StreamType], Any) -> Optional[str]
     """
     Serialize a sequence of Python objects into a YAML stream.
     Produce only basic YAML tags.
@@ -1014,7 +1230,7 @@ def safe_dump_all(documents, stream=None, **kwds):
 
 
 def safe_dump(data, stream=None, **kwds):
-    # type: (Any, Union[None, StreamType], Any) -> Union[None, str]
+    # type: (Any, Optional[StreamType], Any) -> Optional[str]
     """
     Serialize a Python object into a YAML stream.
     Produce only basic YAML tags.
@@ -1043,7 +1259,7 @@ def round_trip_dump(
     top_level_colon_align=None,
     prefix_colon=None,
 ):
-    # type: (Any, Union[None, StreamType], Any, Any, Any, Union[None, bool], Union[None, int], Union[None, int], Union[None, bool], Any, Any, Union[None, bool], Union[None, bool], Union[None, VersionType], Any, Any, Any, Any) -> Union[None, str]   # NOQA
+    # type: (Any, Optional[StreamType], Any, Any, Any, Optional[bool], Optional[int], Optional[int], Optional[bool], Any, Any, Optional[bool], Optional[bool], Optional[VersionType], Any, Any, Any, Any) -> Optional[str]   # NOQA
     allow_unicode = True if allow_unicode is None else allow_unicode
     return dump_all(
         [data],
